/*
 * @Author: renjie.yin
 * @Date: 2022-05-31 19:48:54
 * @LastEditors: renjie.yin
 * @LastEditTime: 2022-06-07 11:07:57
 * @Description: 
 */

// https://github.com/liamappelbe/fftea
// https://github.com/scidart
import 'dart:collection';
import 'dart:math' as math;

import 'fftea.dart';

Map<int, FFT> fftCache = {};

FFT getFFT(int n) {
  if (fftCache.containsKey(n)) {
    var d = fftCache[n];
    if (d != null) {
      return d;
    }
  }
  var d = FFT(n);
  fftCache[n] = d;
  return d;
}

ArrayComplex fftea(List<double> x) {
  var fft = getFFT(x.length);
  var d = fft.realFft(x);
  var res = ArrayComplex.fixed(x.length);

  for (var i = 0; i < x.length; i++) {
    res[i] = Complex.ri(d[i].x, d[i].y);
  }
  return res;
}

const pi = math.pi;

Map<String, Complex> mapCache = {};

Complex complexExpKN(int k, int n) {
  var key = '${k}_$n';
  if (mapCache.containsKey(key)) {
    var d = mapCache[key];
    if (d != null) {
      return d;
    }
  }
  var exponent = Complex.ri(0.0, -2 * pi * k.toDouble() / n.toDouble());

  var d = complexExp(exponent);
  mapCache[key] = Complex.ri(d.real, d.imaginary);
  return d;
}

Complex complexExp(Complex exponent) {
  var a = Complex.ri(math.exp(exponent.real) * math.cos(exponent.imaginary),
      math.exp(exponent.real) * math.sin(exponent.imaginary));
  // var b = Complex.ri(expFast(exponent.real) * math.cos(exponent.imaginary),
  //     expFast(exponent.real) * math.sin(exponent.imaginary));
  return a;
}

Complex complexDivideScalar(Complex a, num b) {
  return Complex(real: a.real / b, imaginary: a.imaginary / b);
}

ArrayComplex arrayComplexDivisionToScalar(ArrayComplex a, num b) {
  var c = ArrayComplex.fixed(a.length);
  for (var i = 0; i < a.length; i++) {
    c[i] = complexDivideScalar(a[i], b);
  }

  return c;
}

bool isEven(int a) => a % 2 == 0;

/// Check if [a] is an Odd number.
bool isOdd(int a) => a % 2 != 0;

/// CHeck if [n] is power of 2.
bool isPowerOf2(int n) => (n & (n - 1)) == 0;

ArrayComplex arrayComplexConcat(ArrayComplex a, ArrayComplex b) {
  var c = ArrayComplex.empty();
  for (var i = 0; i < a.length; i++) {
    c.add(a[i]);
  }
  for (var i = 0; i < b.length; i++) {
    c.add(b[i]);
  }
  return c;
}

Complex complexConjugate(Complex a) {
  return Complex(real: a.real, imaginary: -a.imaginary);
}

class Complex {
  final double real;
  final double imaginary;

  Complex({this.real = 0, this.imaginary = 0});

  Complex.ri(this.real, this.imaginary);

  /// Sum two complex number
  ///
  /// # Examples
  /// ```dart
  /// var c1 = Complex(real: 1, imaginary: 1);
  /// var c2 = Complex(real: 1, imaginary: 1);
  ///
  /// print(c1 + c2);
  /// /* output:
  /// Complex(real: 2.0, imaginary: 2.0);
  /// */
  /// ```
  Complex operator +(Complex b) {
    return Complex(real: real + b.real, imaginary: imaginary + b.imaginary);
  }

  /// Subtract two complex number
  ///
  /// # Examples
  /// ```dart
  /// var c1 = Complex(real: 1, imaginary: 1);
  /// var c2 = Complex(real: 1, imaginary: 1);
  ///
  /// print(c1 - c2);
  ///
  /// /* output:
  /// Complex(real: 0.0, imaginary: 0.0)
  /// */
  /// ```
  Complex operator -(Complex b) {
    return Complex(real: real - b.real, imaginary: imaginary - b.imaginary);
  }

  /// Multiply two complex number
  ///
  /// # Examples
  /// ```dart
  /// var c1 = Complex(real: 2, imaginary: 5);
  /// var c2 = Complex(real: 4, imaginary: -3);
  ///
  /// print(c1 * c2);
  ///
  /// /*output:
  /// Complex(real: 13.0, imaginary: 14.0)
  /// */
  /// ```
  Complex operator *(Complex b) {
    return Complex(
        real: real * b.real - imaginary * b.imaginary,
        imaginary: real * b.imaginary + imaginary * b.real);
  }

  /// Divide two complex number
  ///
  /// # References
  /// 1. "Python Division Of Complex Numbers Without Using Built In Types and Operators". https://stackoverflow.com/a/41146661/6846888. Retrieved 2019-07-15.
  ///
  /// # Examples
  /// ```dart
  /// var c1 = Complex(real: 1, imaginary: -3);
  /// var c2 = Complex(real: 1, imaginary: 2);
  ///
  /// print(c1 / c2);
  ///
  /// /* output:
  /// Complex(real: -1.0, imaginary: -1.0)
  /// */
  /// ```
  Complex operator /(Complex b) {
    var conjugation = complexConjugate(b);
    var denominatorRes = b * conjugation;

    // denominator has only real part
    var denominator = denominatorRes.real;
    var nominator = this * conjugation;

    return Complex(
        real: nominator.real / denominator,
        imaginary: nominator.imaginary / denominator);
  }

  /// Compare two complex number
  ///
  /// # Examples
  /// ```dart
  /// var c1 = Complex(real: 1, imaginary: -3);
  /// var c2 = Complex(real: 1, imaginary: 2);
  ///
  /// print(c1 == c2);
  ///
  /// /* output:
  /// false
  /// */
  @override
  bool operator ==(b) {
    if (b is Complex) {
      return b.real == real && b.imaginary == imaginary;
    } else {
      throw ('the right object has to be a Complex type');
    }
  }

  /// Convert a Complex object (number) to a String representation
  ///
  /// # Examples
  /// ```dart
  /// print(Complex(real: 2.0, imaginary: 2.0));
  ///
  /// var str = Complex(real: 4.0, imaginary: 4.0).toString();
  ///
  /// print(str);
  ///
  /// /* output:
  /// Complex(real: 2.0, imaginary: 2.0)
  /// Complex(real: 4.0, imaginary: 4.0)
  /// */
  /// ```
  @override
  String toString() {
    return '\n    Complex(real: $real, imaginary: $imaginary)';
  }

  double get magnitude {
    return math.sqrt(real * real + imaginary * imaginary).abs();
  }
}

// import 'package:scidart/src/numdart/numdart.dart';

/// Compute the one-dimensional discrete Fourier Transform.
/// Uses recursive Cooley–Tukey algorithm if N is power of 2 otherwise uses
/// Discrete Fourier Transform algorithm.
/// - [x] : An ArrayComplex with the input
/// - [n] : Length of the transformed axis of the output (optional). If not informed,
/// the length of x will be used
/// If n is smaller than the length of the input, the input is cropped.
/// If it is larger, the input is padded with zeros.
/// If n is not given, the length of the input is used.
/// - [normalization] : Compute the FFT normalization wich is: fft(x)/n (optional default false)
/// - return An ArrayComplex with FFT output
///
/// # References
/// 1. "Fast Fourier Transform (C++, Nim and Python examples)". https://rosettacode.org/wiki/Fast_Fourier_transform#Nim. Retrieved 2021-05-10.
/// 2. "Free small FFT in multiple languages". https://www.nayuki.io/page/free-small-fft-in-multiple-languages. Retrieved 2019-07-23.
/// 3. "looking for fft1d arbitrary length code". https://stackoverflow.com/questions/34655959/looking-for-fft1d-arbitrary-length-code. Retrieved 2019-07-23.
/// 4. "Chirp_Z-transform Bluestein". https://en.wikipedia.org/wiki/Chirp_Z-transform#Bluestein.27s_algorithm. Retrieved 2019-07-24.
/// 5. "numpy.fft.fft". https://docs.scipy.org/doc/numpy/reference/generated/numpy.fft.fft.html#numpy.fft.fft. Retrieved 2019-07-24.
/// 6. "Fast fourier transformation polynomial multiplication". https://www.geeksforgeeks.org/fast-fourier-transformation-poynomial-multiplication/. Retrieved 2019-07-25.
/// 7. "Implementation of the Divide and Conquer DFT via Matrices". https://www.projectrhea.org/rhea/index.php/Implementation_of_the_Divide_and_Conquer_DFT_via_Matrices. Retrieved 2019-07-25.
/// 8. "How to implement the discrete fourier transform". https://www.nayuki.io/page/how-to-implement-the-discrete-fourier-transform. Retrieved 2019-07-25.
/// 9. "Fast Fourier transform". https://en.wikipedia.org/wiki/Fast_Fourier_transform. Retrieved 2021-05-10.
/// 10. "Cooley–Tukey FFT algorithm". https://en.wikipedia.org/wiki/Cooley%E2%80%93Tukey_FFT_algorithm#. Retrieved 2021-05-10.
/// 11. "Discrete Fourier transform". https://en.wikipedia.org/wiki/Discrete_Fourier_transform. Retrieved 2021-05-10.
///
/// # Examples
/// ```dart
/// var x = arrayToComplexArray(Array([1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0]));
///
/// fft(x);
///
/// /* output:
/// ArrayComplex([
///   Complex(real: 4.0, imaginary: 0.0),
///   Complex(real: 1.0, imaginary: -2.41421356),
///   Complex(real: 0.0, imaginary: 0.0),
///   Complex(real: 1.0, imaginary: -0.41421356),
///   Complex(real: 0.0, imaginary: 0.0),
///   Complex(real: 1.0, imaginary: 0.41421356),
///   Complex(real: 0.0, imaginary: 0.0),
///   Complex(real: 1.0, imaginary: 2.41421356)
/// ]);
/// */
/// ```

class ArrayComplex extends ListBase<Complex> {
  List<Complex?> l = [];

  //#region constructors
  ArrayComplex(List<Complex> list) {
    l = list.toList();
  }

  ArrayComplex.empty();

  ArrayComplex.fromArray(ArrayComplex list) {
    // deep copy of the parameter
    l = list.map((element) => element).toList();
  }

  ArrayComplex.fixed(int length, {Complex? initialValue}) {
    if (initialValue != null) {
      l = List<Complex>.generate(length, (i) => initialValue);
    } else {
      l = List<Complex>.filled(length, Complex.ri(0.0, 0.0));
    }
  }

  //#endregion

  //#region operators
  @override
  set length(int newLength) {
    if (newLength < 0) {
      throw FormatException(
          'newLength must be greater equal than 0 (newLength >= 0)');
    }

    var aux = l.map((element) => element).toList();

    if (newLength == 0) {
      l = [];
    } else if (newLength > l.length) {
      l = [];
      aux.forEach((element) {
        l.add(element);
      });
      for (var i = l.length; i < newLength; i++) {
        l.add(Complex());
      }
    } else if (newLength < length) {
      l = [];
      for (var i = 0; i < newLength; i++) {
        l.add(aux[i]);
      }
    }
  }

  /// Return the length of ComplexArray
  ///
  /// # Examples
  /// ```dart
  /// var list = ArrayComplex([
  ///   Complex(real: 3.0, imaginary: 4.0),
  ///   Complex(real: 3.0, imaginary: 4.0) ,
  ///   Complex(real: 3.0, imaginary: 4.0)]);
  ///
  /// print(list.length);
  ///
  /// /* output:
  /// 3
  /// */
  /// ```
  @override
  int get length => l.length;

  @override
  Complex operator [](int index) => l[index] ?? Complex();

  @override
  void operator []=(int index, Complex value) {
    l[index] = value;
  }

  /// Multiply two arrays
  ///
  /// # Examples
  /// ```dart
  /// var list = ArrayComplex([
  ///   Complex(real: 2, imaginary: 5),
  ///   Complex(real: 2, imaginary: 5) ,
  ///   Complex(real: 2, imaginary: 5)]);
  /// var list2 = ArrayComplex([
  ///   Complex(real: 4, imaginary: -3),
  ///   Complex(real: 4, imaginary: -3),
  ///   Complex(real: 4, imaginary: -3)]);
  /// var cList = list * list2;
  ///
  /// print(clist)
  ///
  /// /* output:
  /// ArrayComplex([Complex(real: 23.0, imaginary: 14.0), Complex(real: 23.0, imaginary: 14.0), Complex(real: 23.0, imaginary: 14.0)])
  /// */
  /// ```
  ArrayComplex operator *(ArrayComplex b) {
    _checkArray(b);
    var c = ArrayComplex.fixed(length);
    for (var i = 0; i < length; i++) {
      c[i] = this[i] * b[i];
    }

    return c;
  }

  /// compare two arrays
  ///
  /// # Examples
  /// ```dart
  /// var n = Array([1, 2, 3]);
  /// var n2 = Array([1, 2, 3]);
  ///
  /// print(n == n2);
  ///
  /// /* output:
  /// true
  /// */
  /// ``
  @override
  bool operator ==(b) {
    if (b is ArrayComplex) {
      _checkArray(b);
      for (var i = 0; i < length; i++) {
        if (this[i] != b[i]) {
          return false;
        }
      }
      return true;
    } else {
      throw ('the right object has to be an ArrayComplex type');
    }
  }

  /// Return an array given an index interval
  ///
  /// - [start] : start index close interval
  /// - [end] : end index open interval
  /// - [step] : optional, step between samples
  /// - [reverse] : optional, return a reversed array
  ///
  /// # Examples
  /// ```dart
  /// var var list = ArrayComplex([
  ///   Complex(real: 3.0, imaginary: 4.0),
  ///   Complex(real: 3.0, imaginary: 4.0),
  ///   Complex(real: 3.0, imaginary: 4.0)]);
  /// var b = list.getRangeArray(0, 1);
  ///
  /// print(b)
  ///
  /// /* output:
  /// ArrayComplex([
  ///   Complex(real: 3.0, imaginary: 4.0)
  /// ]);
  /// */
  /// ```
  ArrayComplex getRangeArray(int start, int end,
      {int step = 1, bool reverse = false}) {
    var b = ArrayComplex.empty();
    for (var i = start; i < end; i += step) {
      b.add(this[i]);
    }

    if (reverse) {
      b = ArrayComplex(b.reversed.toList());
    }

    return b;
  }
  //#endregion

  //#region overload methods
  /// Convert an ArrayComplex object to a String representation
  ///
  /// # Examples
  /// ```dart
  /// var list = ArrayComplex([
  ///   Complex(real: 3.0, imaginary: 4.0),
  ///   Complex(real: 3.0, imaginary: 4.0) ,
  ///   Complex(real: 3.0, imaginary: 4.0)]);
  ///
  /// print(list.toString());
  /// /* output
  /// ArrayComplex([Complex(real: 3.0, imaginary: 4.0), Complex(real: 3.0, imaginary: 4.0), Complex(real: 3.0, imaginary: 4.0)])
  /// */
  /// ```
  @override
  String toString() {
    var str = IterableBase.iterableToFullString(this, '[', ']');
    return '\n  ArrayComplex($str\n  )';
  }

  //#endregion

  //#region memory operations
  /// Generate a copy of the current vector
  ArrayComplex copy() => ArrayComplex.fromArray(this);

  //#endregion

  //#region private methods
  void _checkArray(ArrayComplex b) {
    if (length != b.length) {
      throw ('both arrays need have the same dimesion');
    }
  }
//#endregion
}

ArrayComplex fft(ArrayComplex x,
    {int? n, bool normalization = false, bool forceDft = false}) {
  var buffer = ArrayComplex.empty();
  if (n == null || n == x.length) {
    n = x.length;
    buffer = x.copy();
  } else if (n < x.length) {
    buffer = ArrayComplex.empty();
    for (var i = 0; i < n; i++) {
      buffer.add(x[i]);
    }
  } else if (n > x.length) {
    buffer = ArrayComplex.empty();
    var pad = n - x.length;
    while (pad > 0) {
      buffer.add(Complex());
      pad--;
    }
    buffer = arrayComplexConcat(buffer, x);
  }

  ArrayComplex res;
  if (buffer.isEmpty) {
    return ArrayComplex.empty();
  } else if (isPowerOf2(n) && !forceDft) {
    // It is power of 2, apply cooley tukey algorithm
    Stopwatch stopwatch = Stopwatch()..start();
    res = _cooleyTukey(buffer);

    stopwatch.stop();
    print('_cooleyTukey() executed in ${stopwatch.elapsed}');
  } else {
    Stopwatch stopwatch = Stopwatch()..start();
    // It is not power of 2, apply discrete fourier transform algorithm
    res = _dft(buffer);
    stopwatch.stop();
    print('_dft() executed in ${stopwatch.elapsed}');
  }

  return normalization ? arrayComplexDivisionToScalar(res, n) : res;
}

ArrayComplex _cooleyTukey(ArrayComplex x) {
  var n = x.length;

  if (n == 0) {
    return ArrayComplex.empty();
  }

  if (n == 1) {
    return ArrayComplex([x[0]]);
  } else {
    var evens = ArrayComplex.empty();
    var odds = ArrayComplex.empty();
    for (var i = 0; i < n; i++) {
      if (isEven(i)) {
        evens.add(x[i]);
      } else {
        odds.add(x[i]);
      }
    }
    var even = _cooleyTukey(evens);
    var odd = _cooleyTukey(odds);

    var result = ArrayComplex.fixed(n);
    var halfn = n ~/ 2;
    for (var k = 0; k < halfn; k++) {
      // var exponent = Complex.ri(0.0, -2 * pi * k.toDouble() / n.toDouble());
      // var a = complexExp(exponent) * odd[k];
      var a = complexExpKN(k, n) * odd[k];
      result[k] = even[k] + a;
      result[k + halfn] = even[k] - a;
    }

    return result;
  }
}

/// Computes the discrete Fourier transform (DFT) of the given complex vector.
/// All the array arguments must be non-null
ArrayComplex _dft(ArrayComplex buffer) {
  var n = buffer.length;
  var out = ArrayComplex.fixed(n);
  for (var k = 0; k < n; k++) {
    // For each output element
    var sumreal = 0.0;
    var sumimag = 0.0;
    for (var t = 0; t < n; t++) {
      // For each input element
      var angle = 2 * math.pi * t * k / n;
      sumreal += buffer[t].real * math.cos(angle) +
          buffer[t].imaginary * math.sin(angle);
      sumimag += -buffer[t].real * math.sin(angle) +
          buffer[t].imaginary * math.cos(angle);
    }
    out[k] = Complex.ri(sumreal, sumimag);
  }

  return out;
}

ArrayComplex arrayComplexConjugate(ArrayComplex a) {
  var c = ArrayComplex.fixed(a.length);
  for (var i = 0; i < a.length; i++) {
    c[i] = complexConjugate(a[i]);
  }
  return c;
}

ArrayComplex ifft(ArrayComplex X) {
  // conjugate the complex numbers
  var XConj = arrayComplexConjugate(X);

  // forward fft
  var xImag = fft(XConj);

  // conjugate the complex numbers again
  xImag = arrayComplexConjugate(xImag);

  // scale the numbers
  xImag = arrayComplexDivisionToScalar(xImag, xImag.length);

  return xImag;
}

extension ComplexEx on Complex {
  Complex div(Complex b1) {
    var a = real;
    var b = imaginary;

    var c = b1.real;
    var d = b1.imaginary;
    var t, x;

    if (0 == d) {
      return Complex(real: a / c, imaginary: b / c);
    }

    if (a.abs() < d.abs()) {
      x = c / d;
      t = c * x + d;

      return Complex(real: (a * x + b) / t, imaginary: (b * x - a) / t);
    } else {
      x = d / c;
      t = d * x + c;

      return Complex(real: (a + b * x) / t, imaginary: (b - a * x) / t);
    }
  }
}

ArrayComplex arrayComplexFromList(List<double> b1) {
  var res = ArrayComplex.empty();
  res.addAll(b1.map((e) => (Complex(real: e))));
  return res;
}

ArrayComplex realFFT(List<double> data, {bool forceDft = false}) {
  if (isPowerOf2(data.length)) {
    Stopwatch stopwatch = Stopwatch()..start();
    var d = fftea(data);
    print('realFFT() executed in ${stopwatch.elapsed}');
    return d;
  }
  Stopwatch stopwatch = Stopwatch()..start();
  var d = arrayComplexFromList(data);

  var fx = fft(d, forceDft: forceDft);
  stopwatch.stop();
  print('realFFT() executed in ${stopwatch.elapsed}');
  return fx;
}

List<double> inverseRealFft(ArrayComplex data) {
  Stopwatch stopwatch = Stopwatch()..start();
  var outData = ifft(data);
  stopwatch.stop();
  print('ifft() executed in ${stopwatch.elapsed}');
  var res = outData.map((item) {
    return item.real;
  }).toList();

  return res;
}
